"""
Unit tests for the aging code.

"""
import os
from datetime import datetime, timedelta
from os.path import join

import pytest
from hdx.database import Database

from hdx.freshness.app.datafreshness import DataFreshness


class TestFreshnessByFrequency:
    @pytest.fixture(scope="class")
    def nodatabase(self):
        dbpath = join("tests", "test_freshness.db")
        try:
            os.remove(dbpath)
        except FileNotFoundError:
            pass
        return {"driver": "sqlite", "database": dbpath}

    @pytest.fixture(scope="class")
    def now(self):
        return datetime.utcnow()

    @pytest.fixture(scope="class")
    def datasets(self):
        return list()

    @pytest.mark.parametrize(
        "days_last_modified,update_frequency,expected_status",
        [
            (0, 1, 0),
            (0, 7, 0),
            (0, 365, 0),
            (1, 1, 1),
            (1, 7, 0),
            (2, 1, 2),
            (3, 1, 3),
            (6, 7, 0),
            (7, 7, 1),
            (8, 7, 1),
            (13, 7, 1),
            (13, 14, 0),
            (14, 7, 2),
            (14, 14, 1),
            (20, 7, 2),
            (20, 14, 1),
            (21, 7, 3),
            (21, 14, 2),
            (29, 14, 3),
            (29, 30, 0),
            (30, 30, 1),
            (30, 90, 0),
            (45, 30, 2),
            (45, 90, 0),
            (60, 30, 3),
            (60, 90, 0),
            (89, 90, 0),
            (90, 90, 1),
            (90, 90, 1),
            (120, 90, 2),
            (150, 90, 3),
            (179, 180, 0),
            (180, 180, 1),
            (210, 180, 2),
            (210, 365, 0),
            (240, 180, 3),
            (240, 365, 0),
            (364, 365, 0),
            (365, 365, 1),
            (425, 365, 2),
            (455, 365, 3),
        ],
    )
    def test_freshness_by_frequency(
        self,
        configuration,
        nodatabase,
        now,
        datasets,
        days_last_modified,
        update_frequency,
        expected_status,
    ):
        with Database(**nodatabase) as session:
            freshness = DataFreshness(
                session=session, datasets=datasets, now=now
            )
            last_modified = now - timedelta(days=days_last_modified)
            status = freshness.calculate_freshness(
                last_modified, update_frequency
            )
            assert status == expected_status



"""
Unit tests for the freshness class.

"""
import json
import logging
import os
import random
from datetime import datetime, timedelta
from os.path import join
from time import sleep

import gspread
import pytest
from gspread.urls import DRIVE_FILES_API_V3_URL
from hdx.api.configuration import Configuration
from hdx.data.dataset import Dataset
from hdx.database import Database

from hdx.freshness.app.datafreshness import DataFreshness
from hdx.freshness.database.dbdataset import DBDataset

logger = logging.getLogger(__name__)


class TestFreshnessCKAN:
    @pytest.fixture(scope="class")
    def configuration(self):
        project_config_yaml = join(
            "src", "hdx", "freshness", "app", "project_configuration.yml"
        )
        hdx_key = os.getenv("HDX_KEY")
        Configuration._create(
            hdx_site="stage",
            user_agent="test",
            hdx_key=hdx_key,
            project_config_yaml=project_config_yaml,
        )

    @pytest.fixture(scope="function")
    def datasetmetadata(self):
        return join("tests", "fixtures", "CKAN", "hdx_dataset_static.yml")

    @pytest.fixture(scope="function")
    def nodatabase(self):
        dbpath = join("tests", "test_freshness.db")
        try:
            os.remove(dbpath)
        except FileNotFoundError:
            pass
        return {"driver": "sqlite", "database": dbpath}

    @pytest.fixture(scope="class")
    def params(self):
        return {
            "corpora": "teamDrive",
            "teamDriveId": "0AKCBfHI3H-hcUk9PVA",
            "supportsAllDrives": True,
            "includeItemsFromAllDrives": True,
        }

    @pytest.fixture(scope="function")
    def gclient(self):
        gsheet_auth = os.getenv("GSHEET_AUTH")
        if not gsheet_auth:
            raise ValueError("No gsheet authorisation supplied!")
        info = json.loads(gsheet_auth)
        scopes = [
            "https://www.googleapis.com/auth/drive",
            "https://www.googleapis.com/auth/spreadsheets",
        ]
        gclient = gspread.service_account_from_dict(info, scopes=scopes)
        return gclient

    @pytest.fixture(scope="function")
    def setup_teardown_folder(self, gclient, params):
        payload = {
            "name": "freshness_test_tmp",
            "mimeType": "application/vnd.google-apps.folder",
            "parents": ["1M8_Hv3myw9RpLq86kBL7QkMAYxcHjvb6"],
        }
        r = gclient.request(
            "post", DRIVE_FILES_API_V3_URL, json=payload, params=params
        )
        folderid = r.json()["id"]
        yield gclient, folderid

        payload = {"trashed": True}
        url = f"{DRIVE_FILES_API_V3_URL}/{folderid}"
        gclient.request("patch", url, json=payload, params=params)

    def test_generate_dataset(
        self,
        configuration,
        datasetmetadata,
        nodatabase,
        setup_teardown_folder,
        params,
    ):
        today = datetime.now()
        gclient, folderid = setup_teardown_folder

        def create_gsheet(name, update):
            payload = {
                "name": name,
                "mimeType": "application/vnd.google-apps.spreadsheet",
                "parents": [folderid],
            }
            r = gclient.request(
                "post", DRIVE_FILES_API_V3_URL, json=payload, params=params
            )
            spreadsheetid = r.json()["id"]
            gsheet = gclient.open_by_key(spreadsheetid)
            wks = gsheet.sheet1
            wks.update("A1", update)
            gsheet.share("", role="reader", perm_type="anyone")
            return wks, f"{gsheet.url}/export?format=xlsx"

        wks, unchanging_url = create_gsheet(
            "unchanging",
            [[random.random() for i in range(4)] for j in range(3)],
        )
        changing_wks1, changing_url1 = create_gsheet(
            "changing1",
            [[random.random() for i in range(5)] for j in range(2)],
        )
        changing_wks2, changing_url2 = create_gsheet(
            "changing2",
            [[random.random() for i in range(3)] for j in range(6)],
        )

        datasets = list()
        last_modifieds = list()
        fresh_dt = datetime.utcnow() - timedelta(days=1)
        due_dt = fresh_dt - timedelta(days=8)
        days7 = timedelta(days=7)
        overdue_dt = due_dt - days7
        delinquent_dt = overdue_dt - days7
        fresh = fresh_dt.isoformat()
        due = due_dt.isoformat()
        overdue = overdue_dt.isoformat()
        delinquent = delinquent_dt.isoformat()
        for i in range(8):
            dataset = Dataset(
                {"name": f"freshness_test_{i}", "title": f"freshness test {i}"}
            )
            dataset.update_from_yaml(datasetmetadata)
            dataset.set_maintainer("196196be-6037-4488-8b71-d786adf4c081")
            dataset.set_organization("5a63012e-6c41-420c-8c33-e84b277fdc90")
            dataset.set_date_of_dataset(today)
            if i == 6:
                dataset.set_expected_update_frequency("Never")
            else:
                dataset.set_expected_update_frequency("Every week")

            dataset.set_subnational(True)
            dataset.add_country_location("AFG")
            tags = ["protests"]
            dataset.add_tags(tags)
            resource = {
                "name": f"test_resource_{i}",
                "description": f"Test Resource {i}",
                "format": "xlsx",
            }
            switcher = {
                0: (unchanging_url, fresh),
                1: (changing_url1, overdue),
                2: (unchanging_url, delinquent),
                3: (unchanging_url, due),
                4: (changing_url2, fresh),
                5: (unchanging_url, overdue),
                6: (unchanging_url, delinquent),
                7: (changing_url1, fresh),
            }
            resource["url"], resource["last_modified"] = switcher.get(i)
            dataset.add_update_resource(resource)
            # add resources
            dataset.create_in_hdx(updated_by_script="freshness_ignore")
            datasets.append(dataset)
            last_modifieds.append({"start": dataset["last_modified"]})
        updated_by_script_dt = None
        try:
            with Database(**nodatabase) as session:
                # first run
                freshness = DataFreshness(
                    session=session, datasets=datasets, do_touch=True
                )
                freshness.spread_datasets()
                freshness.add_new_run()
                hash_ids = [
                    datasets[3].get_resource()["id"],
                    datasets[4].get_resource()["id"],
                    datasets[7].get_resource()["id"],
                ]
                (
                    datasets_to_check,
                    resources_to_check,
                ) = freshness.process_datasets(hash_ids=hash_ids)
                results, hash_results = freshness.check_urls(
                    resources_to_check, "test"
                )
                datasets_lastmodified = freshness.process_results(
                    results, hash_results
                )
                freshness.update_dataset_latest_of_modifieds(
                    datasets_to_check, datasets_lastmodified
                )
                freshness.now.isoformat()
                output1 = freshness.output_counts()

                # change something
                changing_wks1.update(
                    "A1",
                    [[random.random() for i in range(5)] for j in range(2)],
                )
                changing_wks2.update(
                    "A1",
                    [[random.random() for i in range(3)] for j in range(6)],
                )
                sleep(5)

                # second run
                for i, dataset in enumerate(datasets):
                    dataset = Dataset.read_from_hdx(dataset["id"])
                    last_modifieds[i]["run1"] = dataset["last_modified"]
                    if i == 5:
                        dataset["review_date"] = due
                    if i == 7:
                        updated_by_script_dt = datetime.utcnow()
                        updated_by_script = updated_by_script_dt.isoformat()
                        dataset[
                            "updated_by_script"
                        ] = f"freshness ({updated_by_script})"
                    datasets[i] = dataset
                freshness = DataFreshness(
                    session=session, datasets=datasets, do_touch=True
                )
                freshness.spread_datasets()
                freshness.add_new_run()
                (
                    datasets_to_check,
                    resources_to_check,
                ) = freshness.process_datasets(hash_ids=hash_ids)
                results, hash_results = freshness.check_urls(
                    resources_to_check, "test"
                )
                datasets_lastmodified = freshness.process_results(
                    results, hash_results
                )
                freshness.update_dataset_latest_of_modifieds(
                    datasets_to_check, datasets_lastmodified
                )
                run2_last_modified_dt = freshness.now
                run2_last_modified = run2_last_modified_dt.isoformat()
                output2 = freshness.output_counts()
        finally:
            # tear down
            for i, dataset in enumerate(datasets):
                dataset = Dataset.read_from_hdx(dataset["id"])
                if dataset:
                    last_modifieds[i]["run2"] = dataset["last_modified"]
                    dataset.delete_from_hdx()

        assert (
            output1
            == """
*** Resources ***
* total: 8 *,
first hash: 6,
firstrun: 2

*** Datasets ***
* total: 8 *,
0: Fresh, Updated firstrun: 4,
1: Due, Updated firstrun: 1,
2: Overdue, Updated firstrun: 2,
3: Delinquent, Updated firstrun: 1

0 datasets have update frequency of Live
1 datasets have update frequency of Never
0 datasets have update frequency of As Needed"""
        )
        assert (
            output2
            == """
*** Resources ***
* total: 8 *,
hash: 2,
nothing: 3,
same hash: 3

*** Datasets ***
* total: 8 *,
0: Fresh, Updated hash: 2,
0: Fresh, Updated nothing: 2,
0: Fresh, Updated script update: 1,
1: Due, Updated nothing: 1,
1: Due, Updated review date: 1,
3: Delinquent, Updated nothing: 1

0 datasets have update frequency of Live
1 datasets have update frequency of Never
0 datasets have update frequency of As Needed"""
        )

        assert last_modifieds == [
            {"start": fresh, "run1": fresh, "run2": fresh},
            {"start": overdue, "run1": overdue, "run2": run2_last_modified},
            {"start": delinquent, "run1": delinquent, "run2": delinquent},
            {"start": due, "run1": due, "run2": due},
            {"start": fresh, "run1": fresh, "run2": fresh},
            {"start": overdue, "run1": overdue, "run2": overdue},
            {"start": delinquent, "run1": delinquent, "run2": delinquent},
            {"start": fresh, "run1": fresh, "run2": fresh},
        ]
        assert updated_by_script_dt is not None
        expected = [
            {
                "last_modified": fresh_dt,
                "latest_of_modifieds": fresh_dt,
                "what_updated": "nothing",
                "last_resource_modified": fresh_dt,
                "fresh": 0,
            },
            {
                "last_modified": overdue_dt,
                "latest_of_modifieds": run2_last_modified_dt,
                "what_updated": "hash",
                "fresh": 0,
            },
            {
                "last_modified": delinquent_dt,
                "latest_of_modifieds": delinquent_dt,
                "what_updated": "nothing",
                "last_resource_modified": delinquent_dt,
                "fresh": 3,
            },
            {
                "last_modified": due_dt,
                "latest_of_modifieds": due_dt,
                "what_updated": "nothing",
                "last_resource_modified": due_dt,
                "fresh": 1,
            },
            {"last_modified": fresh_dt, "what_updated": "hash", "fresh": 0},
            {
                "review_date": due_dt,
                "last_modified": overdue_dt,
                "latest_of_modifieds": due_dt,
                "what_updated": "review date",
                "last_resource_modified": overdue_dt,
                "fresh": 1,
            },
            {
                "update_frequency": -1,
                "last_modified": delinquent_dt,
                "latest_of_modifieds": delinquent_dt,
                "what_updated": "nothing",
                "last_resource_modified": delinquent_dt,
                "fresh": 0,
            },
            {
                "last_modified": fresh_dt,
                "updated_by_script": updated_by_script_dt,
                "latest_of_modifieds": updated_by_script_dt,
                "what_updated": "script update",
                "last_resource_modified": fresh_dt,
                "fresh": 0,
            },
        ]
        nonmatching = list()
        for i, dataset in enumerate(datasets):
            dbdataset = (
                session.query(DBDataset)
                .filter_by(run_number=1, id=dataset["id"])
                .one()
                .__dict__
            )
            for key, expect in expected[i].items():
                actual = dbdataset[key]
                if actual != expect:
                    nonmatching.append(
                        f"Key {key} of dataset number {i} does not match! {actual} != {expect}"
                    )
        assert nonmatching == list()



"""SQLAlchemy class representing DBDataset row. Holds dynamic dataset metadata for
each run.
"""
from hdx.database import Base
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String

from .dbinfodataset import DBInfoDataset
from .dbrun import DBRun


class DBDataset(Base):
    """
    run_number = Column(
        Integer, ForeignKey(DBRun.run_number), primary_key=True
    )

    id = Column(String, ForeignKey(DBInfoDataset.id), primary_key=True)

    dataset_date = Column(String)

    update_frequency = Column(Integer)

    review_date = Column(DateTime)

    last_modified = Column(DateTime, nullable=False)

    updated_by_script = Column(DateTime)

    metadata_modified = Column(
        DateTime, nullable=False
    )  # this field and above are CKAN fields

    latest_of_modifieds = Column(DateTime, nullable=False)

    what_updated = Column(String, nullable=False)

    last_resource_updated = Column(
        String, nullable=False
    )  # id of last resource updated

    last_resource_modified = Column(
        DateTime, nullable=False
    )  # date last resource updated

    fresh = Column(Integer)

    error = Column(Boolean, nullable=False)
    """

    run_number = Column(
        Integer, ForeignKey(DBRun.run_number), primary_key=True
    )
    id = Column(String, ForeignKey(DBInfoDataset.id), primary_key=True)
    dataset_date = Column(String)
    update_frequency = Column(Integer)
    review_date = Column(DateTime)
    last_modified = Column(DateTime, nullable=False)
    updated_by_script = Column(DateTime)
    metadata_modified = Column(
        DateTime, nullable=False
    )  # this field and above are CKAN fields
    latest_of_modifieds = Column(DateTime, nullable=False)
    what_updated = Column(String, nullable=False)
    last_resource_updated = Column(
        String, nullable=False
    )  # id of last resource updated
    last_resource_modified = Column(
        DateTime, nullable=False
    )  # date last resource updated
    fresh = Column(Integer)
    error = Column(Boolean, nullable=False)

    def __repr__(self) -> str:
        """String representation of DBDataset row

        Returns:
            str: String representation of DBDataset row
        """
        output = f"<Dataset(run number={self.run_number}, id={self.id}, "
        output += f"dataset date={str(self.dataset_date)}, update frequency={self.update_frequency},\n"
        output += f"review date={str(self.review_date)}, last modified={str(self.last_modified)}, "
        output += f"metadata modified={str(self.metadata_modified)}, updated by script={str(self.updated_by_script)},\n"
        output += f"latest of modifieds={str(self.latest_of_modifieds)}, what updated={str(self.what_updated)},\n"
        output += f"Resource {str(self.last_resource_updated)}: last modified={str(self.last_resource_modified)},\n"
        output += f"Dataset fresh={str(self.fresh)}, error={str(self.error)})>"
        return output



"""SQLAlchemy class representing DBTestResult row. Holds test data mimicking the result
of downloading and hashing urls (second time).
"""
from hdx.database import Base
from sqlalchemy import Boolean, Column, DateTime, String


class DBTestHashResult(Base):
    """
    id = Column(String, primary_key=True)
    url = Column(String, nullable=False)
    format = Column(String, nullable=False)
    err = Column(String)
    http_last_modified = Column(DateTime)
    hash = Column(String)
    xlsx_hash = Column(String)
    force_hash = Column(Boolean, nullable=False)
    """

    id = Column(String, primary_key=True)
    url = Column(String, nullable=False)
    format = Column(String, nullable=False)
    err = Column(String)
    http_last_modified = Column(DateTime)
    hash = Column(String)
    xlsx_hash = Column(String)
    force_hash = Column(Boolean, nullable=False)

    def __repr__(self) -> str:
        """String representation of DBTestHashresult row

        Returns:
            str: String representation of DBTestHashResult row
        """
        output = f"<TestHashResult(id={self.id}, url={self.url}, "
        output += f"format={self.format}, err={self.err}\n"
        output += f"http_last_modified={str(self.http_last_modified)}, "
        output += f"hash={self.hash}, xlsx_hash={self.xlsx_hash}, "
        output += f"force_hash={str(self.force_hash)})>"
        return output



"""SQLAlchemy class representing DBTestResource row. Holds test data for resources with
the aim of mimicking what would come from HDX.
"""
from hdx.database import Base
from sqlalchemy import Column, ForeignKey, String

from .dbtestdataset import DBTestDataset


class DBTestResource(Base):
    """
    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    format = Column(String, nullable=False)
    dataset_id = Column(String, ForeignKey(DBTestDataset.id), nullable=False)
    url = Column(String, nullable=False)
    metadata_modified = Column(String, nullable=False)
    last_modified = Column(String, nullable=False)
    """

    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    format = Column(String, nullable=False)
    dataset_id = Column(String, ForeignKey(DBTestDataset.id), nullable=False)
    url = Column(String, nullable=False)
    metadata_modified = Column(String, nullable=False)
    last_modified = Column(String, nullable=False)

    def __repr__(self) -> str:
        """String representation of DBTestResource row

        Returns:
            str: String representation of DBTestResource row
        """
        output = f"<Resource(id={self.id}, name={self.name},\n"
        output += f"format={self.format}, dataset id={self.dataset_id},\n"
        output += f"url={self.url},\n"
        output += f"metadata modified={str(self.metadata_modified)}, "
        output += f"last_modified={str(self.last_modified)})>"
        return output



"""SQLAlchemy class representing DBRun row. Holds date of each run.
"""
from hdx.database import Base
from sqlalchemy import Column, DateTime, Integer


class DBRun(Base):
    """
    run_number = Column(Integer, primary_key=True)
    run_date = Column(DateTime)
    """

    run_number = Column(Integer, primary_key=True)
    run_date = Column(DateTime)

    def __repr__(self) -> str:
        """String representation of DBRun row

        Returns:
            str: String representation of DBRun row
        """
        return f"<Run number={self.run_number}, Run date={str(self.run_date)}>"



"""Utility to download and hash resources. Uses asyncio. Note that the purpose of
asyncio is to help with IO-bound rather than CPU-bound code (for which multiprocessing
is more suitable as it leverages multiple CPUs). Asyncio allows you to structure your
code so that when one piece of linear single-threaded code (coroutine) is waiting for
something to happen another can take over and use the CPU. While conceptually similar to
threading, the difference is that with asyncio, it is the task of the developer rather
than the OS to decide when to switch to the next task.
"""
import asyncio
import hashlib
import logging
from io import BytesIO
from timeit import default_timer as timer
from typing import Dict, List, Optional, Tuple, Union

import aiohttp
import tqdm
import uvloop
from dateutil import parser
from openpyxl import load_workbook

from . import retry
from .ratelimiter import RateLimiter

logger = logging.getLogger(__name__)


class Retrieval:
    """Retrieval class for downloading and hashing resources.

    Args:
        user_agent (str): User agent string to use when downloading
        url_ignore (Optional[str]): Parts of url to ignore for special xlsx handling
    """

    ignore_mimetypes = ["application/octet-stream", "application/binary"]
    mimetypes = {
        "json": ["application/json"],
        "geojson": ["application/json", "application/geo+json"],
        "shp": ["application/zip", "application/x-zip-compressed"],
        "csv": ["text/csv", "application/zip", "application/x-zip-compressed"],
        "xls": ["application/vnd.ms-excel"],
        "xlsx": [
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
        ],
    }
    signatures = {
        "json": [b"[", b" [", b"{", b" {"],
        "geojson": [b"[", b" [", b"{", b" {"],
        "shp": [b"PK\x03\x04"],
        "xls": [b"\xd0\xcf\x11\xe0"],
        "xlsx": [b"PK\x03\x04"],
    }

    def __init__(
        self, user_agent: str, url_ignore: Optional[str] = None
    ) -> None:
        self.user_agent = user_agent
        self.url_ignore: Optional[str] = url_ignore

    async def fetch(
        self,
        metadata: Tuple,
        session: Union[aiohttp.ClientSession, RateLimiter],
    ) -> Tuple:
        """Asynchronous code to download a resource and hash it. Returns a tuple with
        resource information including hashes.

        Args:
            metadata (Tuple): Resource to be checked
            session (Union[aiohttp.ClientSession, RateLimiter]): session to use for requests

        Returns:
            Tuple: Resource information including hash
        """
        url = metadata[0]
        resource_id = metadata[1]
        resource_format = metadata[2]

        async def fn(response):
            last_modified_str = response.headers.get("Last-Modified")
            http_last_modified = None
            if last_modified_str:
                try:
                    # we set http_last_modified but don't actually use it to calculate
                    # freshness any more
                    http_last_modified = parser.parse(
                        last_modified_str, ignoretz=True
                    )
                except (ValueError, OverflowError):
                    pass
            length = response.headers.get("Content-Length")
            if length and int(length) > 104857600:
                response.close()
                err = "File too large to hash!"
                return (
                    resource_id,
                    url,
                    resource_format,
                    err,
                    http_last_modified,
                    None,
                    None,
                )
            logger.info(f"Hashing {url}")
            mimetype = response.headers.get("Content-Type")

            try:
                iterator = response.content.iter_any()
                first_chunk = await iterator.__anext__()
                signature = first_chunk[:4]
                if (
                    resource_format == "xlsx"
                    and mimetype == self.mimetypes["xlsx"][0]
                    and signature == self.signatures["xlsx"][0]
                    and (
                        self.url_ignore not in url if self.url_ignore else True
                    )
                ):
                    xlsxbuffer = bytearray(first_chunk)
                else:
                    xlsxbuffer = None
                md5hash = hashlib.md5(first_chunk)
                async for chunk in iterator:
                    if chunk:
                        md5hash.update(chunk)
                        if xlsxbuffer:
                            xlsxbuffer.extend(chunk)
                if xlsxbuffer:
                    workbook = load_workbook(filename=BytesIO(xlsxbuffer))
                    xlsx_md5hash = hashlib.md5()
                    for sheet_name in workbook.sheetnames:
                        sheet = workbook[sheet_name]
                        for cols in sheet.iter_rows(values_only=True):
                            xlsx_md5hash.update(bytes(str(cols), "utf-8"))
                    xlsxbuffer = None
                else:
                    xlsx_md5hash = None
                err = None
                if mimetype not in self.ignore_mimetypes:
                    expected_mimetypes = self.mimetypes.get(resource_format)
                    if expected_mimetypes is not None:
                        if not any(x in mimetype for x in expected_mimetypes):
                            err = f"File mimetype {mimetype} does not match HDX format {resource_format}!"
                expected_signatures = self.signatures.get(resource_format)
                if expected_signatures is not None:
                    found = False
                    for expected_signature in expected_signatures:
                        if (
                            signature[: len(expected_signature)]
                            == expected_signature
                        ):
                            found = True
                            break
                    if not found:
                        sigerr = f"File signature {signature} does not match HDX format {resource_format}!"
                        if err is None:
                            err = sigerr
                        else:
                            err = f"{err} {sigerr}"
                return (
                    resource_id,
                    url,
                    resource_format,
                    err,
                    http_last_modified,
                    md5hash.hexdigest(),
                    xlsx_md5hash.hexdigest() if xlsx_md5hash else None,
                )
            except Exception as exc:
                try:
                    code = exc.code
                except AttributeError:
                    code = ""
                err = f"Exception during hashing: code={code} message={exc} raised={exc.__class__.__module__}.{exc.__class__.__qualname__} url={url}"
                raise aiohttp.ClientResponseError(
                    code=code,
                    message=err,
                    request_info=response.request_info,
                    history=response.history,
                ) from exc

        try:
            return await retry.send_http(
                session, "get", url, retries=2, interval=5, backoff=4, fn=fn
            )
        except Exception as e:
            return resource_id, url, resource_format, str(e), None, None, None

    async def check_urls(
        self, resources_to_check: List[Tuple], loop: uvloop.Loop
    ) -> Dict[str, Tuple]:
        """Asynchronous code to download resources and hash them. Return dictionary with
        resources information including hashes.

        Args:
            resources_to_check (List[Tuple]): List of resources to be checked
            loop (uvloop.Loop): Event loop to use

        Returns:
            Dict[str, Tuple]: Resources information including hashes
        """
        tasks = list()

        conn = aiohttp.TCPConnector(limit=100, limit_per_host=1, loop=loop)
        timeout = aiohttp.ClientTimeout(
            total=60 * 60, sock_connect=30, sock_read=30
        )
        async with aiohttp.ClientSession(
            connector=conn,
            timeout=timeout,
            loop=loop,
            headers={"User-Agent": self.user_agent},
        ) as session:
            session = RateLimiter(
                session
            )  # Limit connections per timeframe to host
            for metadata in resources_to_check:
                task = self.fetch(metadata, session)
                tasks.append(task)
            responses = dict()
            for f in tqdm.tqdm(asyncio.as_completed(tasks), total=len(tasks)):
                (
                    resource_id,
                    url,
                    resource_format,
                    err,
                    http_last_modified,
                    hash,
                    hash_xlsx,
                ) = await f
                responses[resource_id] = (
                    url,
                    resource_format,
                    err,
                    http_last_modified,
                    hash,
                    hash_xlsx,
                )
            return responses

    def retrieve(self, resources_to_check: List[Tuple]) -> Dict[str, Tuple]:
        """Download resources and hash them. Return dictionary with resources information
        including hashes.

        Args:
            resources_to_check (List[Tuple]): List of resources to be checked

        Returns:
            Dict[str, Tuple]: Resources information including hashes
        """

        start_time = timer()
        loop = uvloop.new_event_loop()
        asyncio.set_event_loop(loop)
        future = asyncio.ensure_future(
            self.check_urls(resources_to_check, loop)
        )
        results = loop.run_until_complete(future)
        logger.info(f"Execution time: {timer() - start_time} seconds")
        loop.run_until_complete(asyncio.sleep(0.250))
        loop.close()
        return results



"""SQLAlchemy class representing DBOrganization row. Holds static organisation metadata.
"""
from hdx.database import Base
from sqlalchemy import Column, String


class DBOrganization(Base):
    """
    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    title = Column(String, nullable=False)
    """

    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    title = Column(String, nullable=False)

    def __repr__(self) -> str:
        """String representation of DBOrganization row

        Returns:
            str: String representation of DBOrganization row
        """
        return f"<Organization(id={self.id}, name={self.name}, title={self.title})>"



"""
Unit tests for process_results.

"""
import datetime
from unittest.mock import Mock

import pytest
from dateutil import parser
from hdx.data.dataset import Dataset

from hdx.freshness.app.datafreshness import DataFreshness


class TestProcessResults:
    @pytest.fixture(scope="function")
    def session(self):
        class TestSession:
            @staticmethod
            def query(somethingin):
                something = str(somethingin)
                result = Mock()
                result.scalar.return_value = False
                result.distinct.return_value.order_by.return_value.first.return_value = (
                    None
                )

                if "DBResource" in something:

                    class DBResource:
                        dataset_id = "c1c85ecb-5e84-48c6-8ba9-15689a6c2fc4"
                        what_updated = ""
                        http_last_modified = None
                        latest_of_modifieds = parser.parse(
                            "2019-10-28 05:05:20"
                        )
                        md5_hash = "5600bafa19852afae3d7fd27955df0e6"
                        error = ""

                    result.filter_by.return_value.one.return_value = (
                        DBResource()
                    )
                else:

                    class DBDataset:
                        fresh = 0
                        update_frequency = 7

                    result.filter_by.return_value.one.return_value = (
                        DBDataset()
                    )
                return result

            @staticmethod
            def commit():
                pass

        return TestSession()

    @pytest.fixture(scope="function")
    def now(self):
        return parser.parse("2019-11-03 23:01:31.438713")

    @pytest.fixture(scope="function")
    def datasets(self):
        ds = {
            "title": "HOTOSM Afghanistan Points of Interest (OpenStreetMap Export)",
            "package_creator": "osm2hdx",
            "data_update_frequency": "7",
            "maintainer": "6a0688ce-8521-46e2-8edd-8e26c0851ebd",
            "private": False,
            "dataset_date": "11/03/2019",
            "id": "c1c85ecb-5e84-48c6-8ba9-15689a6c2fc4",
            "metadata_created": "2019-08-16T00:35:57.276354",
            "metadata_modified": "2019-11-04T08:45:39.230224",
            "subnational": "1",
            "methodology": "Other",
            "license_id": "hdx-odc-odbl",
            "dataset_source": "OpenStreetMap contributors",
            "tags": [
                {
                    "vocabulary_id": "b891512e-9516-4bf5-962a-7a289772a2a1",
                    "name": "facilities and infrastructure",
                },
                {
                    "vocabulary_id": "b891512e-9516-4bf5-962a-7a289772a2a1",
                    "name": "points of interest - poi",
                },
            ],
            "last_modified": "2019-08-16T00:48:44.310234",
            "groups": [{"name": "afg"}],
            "methodology_other": "Volunteered geographic information",
            "name": "hotosm_afg_points_of_interest",
            "notes": "OpenStreetMap",
            "owner_org": "225b9f7d-e7cb-4156-96a6-44c9c58d31e3",
            "resources": [
                {
                    "id": "fe295e74-56d6-4a83-b50c-43001a204b0f",
                    "metadata_modified": "2019-11-04T08:45:39.226896",
                    "description": "ESRI Shapefile",
                    "format": "SHP",
                    "last_modified": "2019-08-16T00:48:38.609336",
                    "name": "hotosm_afg_points_of_interest_points_shp.zip",
                    "created": "2019-08-16T00:48:38.783589",
                    "url": "http://export.hotosm.org/downloads/1364e367-304e-4df2-989c-839760c3728d/hotosm_afg_points_of_interest_points_shp.zip",
                },
                {
                    "id": "bd64f67f-aa13-479d-85a6-c05f036a210c",
                    "metadata_modified": "2019-11-04T08:45:39.226896",
                    "description": "ESRI Shapefile",
                    "format": "SHP",
                    "last_modified": "2019-08-16T00:48:38.609336",
                    "name": "hotosm_afg_points_of_interest_polygons_shp.zip",
                    "created": "2019-08-16T00:48:38.783599",
                    "url": "http://export.hotosm.org/downloads/1364e367-304e-4df2-989c-839760c3728d/hotosm_afg_points_of_interest_polygons_shp.zip",
                },
                {
                    "id": "dc8be6ea-d9d5-48e0-afe1-7e5f7e6be652",
                    "metadata_modified": "2019-11-04T08:45:39.226896",
                    "description": "Geopackage, SQLite compatible",
                    "format": "Geopackage",
                    "last_modified": "2019-08-16T00:48:44.310234",
                    "name": "hotosm_afg_points_of_interest_gpkg.zip",
                    "created": "2019-08-16T00:48:38.783604",
                    "url": "http://export.hotosm.org/downloads/1364e367-304e-4df2-989c-839760c3728d/hotosm_afg_points_of_interest_gpkg.zip",
                },
                {
                    "id": "5df64cf7-d782-47f2-99a4-5258f1389033",
                    "metadata_modified": "2019-11-04T08:45:39.226896",
                    "description": "Google Earth .KML",
                    "format": "KML",
                    "last_modified": "2019-08-16T00:48:44.310234",
                    "name": "hotosm_afg_points_of_interest_points_kml.zip",
                    "created": "2019-08-16T00:48:38.783608",
                    "url": "http://export.hotosm.org/downloads/1364e367-304e-4df2-989c-839760c3728d/hotosm_afg_points_of_interest_points_kml.zip",
                },
                {
                    "id": "3adb573a-f056-41b7-8ee5-ec245676a7ce",
                    "metadata_modified": "2019-11-04T08:45:39.226896",
                    "description": "Google Earth .KML",
                    "format": "KML",
                    "last_modified": "2019-08-16T00:48:44.310234",
                    "name": "hotosm_afg_points_of_interest_polygons_kml.zip",
                    "created": "2019-08-16T00:48:38.783613",
                    "url": "http://export.hotosm.org/downloads/1364e367-304e-4df2-989c-839760c3728d/hotosm_afg_points_of_interest_polygons_kml.zip",
                },
                {
                    "id": "5cf4261f-b571-4bf4-9a5c-2998f49be722",
                    "metadata_modified": "2019-11-04T08:45:39.226896",
                    "description": ".IMG for Garmin GPS Devices (All OSM layers for area)",
                    "format": "ERDAS Image",
                    "last_modified": "2019-08-16T00:48:44.310234",
                    "name": "hotosm_afg_gmapsupp_img.zip",
                    "created": "2019-08-16T00:48:38.783617",
                    "url": "http://export.hotosm.org/downloads/1364e367-304e-4df2-989c-839760c3728d/hotosm_afg_gmapsupp_img.zip",
                },
            ],
        }
        return [Dataset(ds)]

    @pytest.fixture(scope="function")
    def results(self):
        results = {
            "3adb573a-f056-41b7-8ee5-ec245676a7ce": (
                "http://export.hotosm.org/downloads/1364e367-304e-4df2-989c-839760c3728d/hotosm_afg_points_of_interest_polygons_kml.zip",
                "application/zip",
                None,
                parser.parse("2019-11-03 14:23:40"),
                "33caf1b1106613d123989c2b459c383d",
                None,
            )
        }
        return results

    def test_process_results(
        self, configuration, session, now, datasets, results, resourcecls
    ):
        freshness = DataFreshness(
            session=session, datasets=datasets, now=now, do_touch=True
        )
        resourcecls.populate_resourcedict(datasets)
        resourcecls.touched = False
        datasets_lastmodified = freshness.process_results(
            results, results, resourcecls=resourcecls
        )
        assert datasets_lastmodified == {
            "c1c85ecb-5e84-48c6-8ba9-15689a6c2fc4": {
                "3adb573a-f056-41b7-8ee5-ec245676a7ce": (
                    "",
                    datetime.datetime(2019, 11, 3, 23, 1, 31, 438713),
                    ",hash",
                )
            }
        }
        assert resourcecls.touched is True



"""
Unit tests for the serializing code.

"""
import datetime
import os
import pickle
from os.path import join

import pytest
from hdx.database import Database

from hdx.freshness.testdata.serialize import (
    deserialize_datasets,
    deserialize_hashresults,
    deserialize_now,
    deserialize_results,
    serialize_datasets,
    serialize_hashresults,
    serialize_now,
    serialize_results,
)


class TestSerialize:
    @pytest.fixture(scope="function")
    def session(self):
        dbpath = join("tests", "test_serialize.db")
        try:
            os.remove(dbpath)
        except FileNotFoundError:
            pass
        return Database.get_session(f"sqlite:///{dbpath}")

    @pytest.fixture(scope="function")
    def datasets(self):
        fixture = join("tests", "fixtures", "datasets.pickle")
        with open(fixture, "rb") as fp:
            return pickle.load(fp)

    @staticmethod
    def get_deserialized(d):
        return {
            key: (value[0], value[1], value[2], value[3], value[4], value[5])
            for key, value in d.items()
        }

    def test_serialize_datasets(self, configuration, session, datasets):
        serialize_datasets(session, datasets)
        for i, result in enumerate(deserialize_datasets(session)):
            dataset = datasets[i]
            assert (
                result["organization"]["id"] == dataset["organization"]["id"]
            )
            assert (
                result["organization"]["name"]
                == dataset["organization"]["name"]
            )
            assert (
                result["organization"]["title"]
                == dataset["organization"]["title"]
            )
            assert result["name"] == dataset["name"]
            assert result["title"] == dataset["title"]
            assert result["private"] == dataset["private"]
            assert result["maintainer"] == dataset["maintainer"]
            assert result["dataset_date"] == dataset.get("dataset_date")
            assert result["metadata_modified"] == dataset["metadata_modified"]
            assert result["review_date"] == dataset["review_date"]
            assert result["last_modified"] == dataset["last_modified"]
            assert result["data_update_frequency"] == dataset.get(
                "data_update_frequency"
            )
            assert result["is_requestdata_type"] == dataset.get(
                "is_requestdata_type"
            )
            assert result["groups"] == [
                {"name": x["name"]} for x in dataset.get("groups")
            ]
            resources = dataset.get_resources()
            for j, result_resource in enumerate(result.get_resources()):
                resource = resources[j]
                assert result_resource["id"] == resource["id"]
                assert result_resource["name"] == resource["name"]
                assert result_resource["format"] == resource["format"]
                assert result_resource["url"] == resource["url"]
                assert (
                    result_resource["metadata_modified"]
                    == resource["metadata_modified"]
                )
                assert (
                    result_resource["last_modified"]
                    == resource["last_modified"]
                )

    def test_serialize_now(self, session):
        now = datetime.datetime.utcnow()
        serialize_now(session, now)
        result = deserialize_now(session)
        assert result == now

    def test_serialize_results(self, session):
        results = {
            "c001c17e-92e1-4c15-ac7f-e9a2a22bc11e": (
                "http://siged.sep.gob.mx/SIGED/content/conn/WCPortalUCM/path/Contribution%20Folders/PortalSIGED/Descargas/Datos%20Abiertos/Censo/Personal/Cuestionarios/PERSONAL_CUES_CT_CENSADOS.zip",
                "application/zip",
                "code=404 message=Non-retryable response code raised=aiohttp.errors.HttpProcessingError url=http://siged.sep.gob.mx/SIGED/content/conn/WCPortalUCM/path/Contribution%20Folders/PortalSIGED/Descargas/Datos%20Abiertos/Censo/Personal/Cuestionarios/PERSONAL_CUES_CT_CENSADOS.zip",
                None,
                None,
                None,
                False,
            ),
            "563e2bd1-b200-416b-99be-425777ad686a": (
                "http://geonode.state.gov/geoserver/wms/kml?layers=geonode%3ASyria_BorderCrossings_2015Jun11_HIU_USDoS&mode=download",
                "application/zip",
                None,
                None,
                "bde2adc82876bd845cc4c5233c224a10",
                None,
                False,
            ),
            "a9cb1b9e-93b2-4ff4-82a7-3aab8b13d7b6": (
                "http://www.majidata.go.ke/dataset_dl.php?meza=H_County_WaterSupply_ALinked",
                "application/zip",
                None,
                None,
                "1a1e0af350ac825ba21adefd94926c9d",
                None,
                False,
            ),
        }
        serialize_results(session, results)
        result = deserialize_results(session)
        assert result == self.get_deserialized(results)

    def test_serialize_hash_results(self, session):
        hash_results = {
            "dc8da7da-59bc-4fad-98b8-9a0303b2deed": (
                "http://data.humdata.org/dataset/e66dbc70-17fe-4230-b9d6-855d192fc05c/resource/dc8da7da-59bc-4fad-98b8-9a0303b2deed/download/myanmar-adm2.geojson",
                "application/json",
                None,
                datetime.datetime(2015, 7, 24, 7, 8, 48),
                "73ba2b7904c778ed218357d9c1515c0c",
                None,
                True,
            ),
            "3eb2c0ac-4b27-49b6-be25-f5ccb7128d65": (
                "http://sddr.faoswalim.org/Shapefiles/Administrative/Somalia%20Major%20Primary%20Roads.ZIP",
                "application/zip",
                None,
                datetime.datetime(2013, 2, 28, 13, 53, 42),
                "3e59e8be4973de25eaa4283e075ad5b2",
                None,
                True,
            ),
            "e3eea5de-80bf-4b2a-9729-31b89d6fb36c": (
                "http://ourairports.com/countries/RS/airports.hxl",
                "text/csv",
                None,
                None,
                "71d1ecb069dbd2fc32f79eb6e0859c55",
                None,
                True,
            ),
            "e351d04f-fade-45f9-81fa-0ea673bd9b33": (
                "https://docs.google.com/spreadsheets/d/1kPO1CmPvc42j9TouovP5tyiSdegRk1wISLpxMWDlfaQ/pub?gid=0&single=true&output=csv",
                "text/csv",
                None,
                None,
                "6827a97f982da889840d03f568a04f32",
                None,
                False,
            ),
        }
        serialize_hashresults(session, hash_results)
        result = deserialize_hashresults(session)
        assert result == self.get_deserialized(hash_results)



"""aiohttp rate limiting: limit connections per timeframe to host
(from https://quentin.pradet.me/blog/how-do-you-rate-limit-calls-with-aiohttp.html)
"""
import asyncio
import time
from typing import Any
from urllib.parse import urlsplit

from aiohttp.client import _RequestContextManager


class RateLimiter:
    """
    Use like this:
    session = RateLimiter(session)
    ...
    async with await session.get(url) as response

    Args:
        session (aiohttp.ClientSession): aiohttp session to use for requests
    """

    RATE = 9 / 60  # requests per second
    MAX_TOKENS = 10

    def __init__(self, session):
        self.session = session
        self.start_time = time.monotonic()
        self.tokens = dict()

    async def get(
        self, url: str, *args: Any, **kwargs: Any
    ) -> _RequestContextManager:
        """Asynchronous code to download a resource after waiting for a token

        Args:
            url (str): Url to download
            *args
            **kwargs

        Returns:
            aiohttp.ClientResponse: Response from request
        """
        host = urlsplit(url).netloc
        await self.wait_for_token(host)
        return self.session.get(url, *args, **kwargs)

    async def wait_for_token(self, host: str) -> None:
        """Asynchronous code to handle sleeping if host already connected to

        Args:
            host (str): Host (server)

        Returns:
            None
        """
        if host not in self.tokens:
            self.tokens[host] = [self.MAX_TOKENS, self.start_time]
        while self.tokens[host][0] < 1:
            self.add_new_tokens(host)
            await asyncio.sleep(1)
        self.tokens[host][0] -= 1

    def add_new_tokens(self, host: str) -> None:
        """Adds new tokens

        Args:
            host (str): Host (server)

        Returns:
            None
        """
        now = time.monotonic()
        time_since_update = now - self.tokens[host][1]
        new_tokens = time_since_update * self.RATE
        if new_tokens > 1:
            self.tokens[host][0] = min(
                self.tokens[host][0] + new_tokens, self.MAX_TOKENS
            )
            self.tokens[host][1] = now



"""SQLAlchemy class representing DBInfoDataset row. Holds static dataset metadata.
"""
from hdx.database import Base
from sqlalchemy import Boolean, Column, ForeignKey, String

from .dborganization import DBOrganization


class DBInfoDataset(Base):
    """
    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    title = Column(String, nullable=False)
    private = Column(Boolean, nullable=False)
    organization_id = Column(
        String, ForeignKey(DBOrganization.id), nullable=False
    )
    location = Column(String)
    maintainer = Column(String)
    """

    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    title = Column(String, nullable=False)
    private = Column(Boolean, nullable=False)
    organization_id = Column(
        String, ForeignKey(DBOrganization.id), nullable=False
    )
    location = Column(String)
    maintainer = Column(String)

    def __repr__(self) -> str:
        """String representation of DBInfoDataset row

        Returns:
            str: String representation of DBInfoDataset row
        """
        output = f"<InfoDataset(id={self.id}, name={self.name}, title={self.title},\n"
        output += f"private={str(self.private)}, organization id={self.organization_id},\n"
        output += f"maintainer={self.maintainer}, location={self.location})>"
        return output



"""SQLAlchemy class representing DBResource row. Holds dynamic resource metadata for
each run.
"""
from hdx.database import Base
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String

from .dbinfodataset import DBInfoDataset
from .dbrun import DBRun


class DBResource(Base):
    """
    run_number = Column(
        Integer, ForeignKey(DBRun.run_number), primary_key=True
    )
    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    dataset_id = Column(String, ForeignKey(DBInfoDataset.id), nullable=False)
    url = Column(String, nullable=False)
    last_modified = Column(DateTime, nullable=False)
    metadata_modified = Column(
        DateTime, default=None
    )  # this field and above are CKAN fields
    latest_of_modifieds = Column(DateTime, nullable=False)
    what_updated = Column(String, nullable=False)
    http_last_modified = Column(DateTime, default=None)
    md5_hash = Column(String, default=None)
    hash_last_modified = Column(DateTime, default=None)
    when_checked = Column(DateTime, default=None)
    api = Column(Boolean)
    error = Column(String)
    """

    run_number = Column(
        Integer, ForeignKey(DBRun.run_number), primary_key=True
    )
    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    dataset_id = Column(String, ForeignKey(DBInfoDataset.id), nullable=False)
    url = Column(String, nullable=False)
    last_modified = Column(DateTime, nullable=False)
    metadata_modified = Column(
        DateTime, default=None
    )  # this field and above are CKAN fields
    latest_of_modifieds = Column(DateTime, nullable=False)
    what_updated = Column(String, nullable=False)
    http_last_modified = Column(DateTime, default=None)
    md5_hash = Column(String, default=None)
    hash_last_modified = Column(DateTime, default=None)
    when_checked = Column(DateTime, default=None)
    api = Column(Boolean)
    error = Column(String)

    def __repr__(self) -> str:
        """String representation of DBResource row

        Returns:
            str: String representation of DBResource row
        """
        output = f"<Resource(run number={self.run_number}, id={self.id}, name={self.name}, "
        output += f"dataset id={self.dataset_id},\nurl={self.url},\n"
        output += f"last modified={str(self.last_modified)}, metadata modified={str(self.metadata_modified)},\n"
        output += f"latest of modifieds={str(self.latest_of_modifieds)}, what updated={str(self.what_updated)},\n"
        output += f"http last modified={str(self.http_last_modified)},\n"
        output += f"MD5 hash={self.md5_hash}, hash last modified={str(self.hash_last_modified)}, "
        output += f"when checked={str(self.when_checked)},\n"
        output += f"api={str(self.api)}, error={str(self.error)})>"
        return output
